==> jupiter-2.3.0-1.21.1-neoforge-sources.jar: com/iafenvoy/jupiter/compat/forgeconfigspec/NightConfigHolder.java <==
package com.iafenvoy.jupiter.compat.forgeconfigspec;

import com.electronwill.nightconfig.core.CommentedConfig;
import com.electronwill.nightconfig.core.UnmodifiableConfig;
import com.iafenvoy.jupiter.Jupiter;
import com.iafenvoy.jupiter.config.ConfigGroup;
import com.iafenvoy.jupiter.config.ConfigSide;
import com.iafenvoy.jupiter.config.entry.*;
import com.iafenvoy.jupiter.config.interfaces.ConfigBuilder;
import com.iafenvoy.jupiter.interfaces.IConfigEntry;
import com.iafenvoy.jupiter.util.TextFormatter;
import com.iafenvoy.jupiter.util.TextUtil;
import net.minecraft.network.chat.Component;
import net.minecraft.resources.ResourceLocation;
//? >= 1.20.2 {
import net.neoforged.neoforge.common.ModConfigSpec;
//?} else {
/*import net.minecraftforge.common.ForgeConfigSpec;
 *///?}

import java.util.*;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiFunction;
import java.util.function.Predicate;

//WARNING!!! DO NOT try to understand how these code work!!!
public final class NightConfigHolder {
    private final String modId;
    private final ConfigSide side;
    private final String fileName;
    private final UnmodifiableConfig defaults;
    private final CommentedConfig values;
    private final Runnable save;

    public NightConfigHolder(String modId, ConfigSide side, String fileName, UnmodifiableConfig defaults, CommentedConfig values, Runnable save) {
        this.modId = modId;
        this.side = side;
        this.fileName = fileName;
        this.defaults = defaults;
        this.values = values;
        this.save = save;
    }

    public ResourceLocation id() {
        return Jupiter.id(this.modId, this.side.name().toLowerCase(Locale.ROOT));
    }

    public Component title() {
        return TextUtil.literal(TextFormatter.formatToTitleCase(this.modId + "_" + this.side.name().toLowerCase(Locale.ROOT) + "_config"));
    }

    public ConfigSide getSide() {
        return this.side;
    }

    public String getFileName() {
        return this.fileName;
    }

    public void save() {
        this.save.run();
    }

    public List<ConfigGroup> toGroups() {
        return List.of(this.buildGroup(this.id().toString(), this.title(), this.defaults, this.values));
    }

    public ConfigGroup buildGroup(String id, Component groupName, UnmodifiableConfig defaults, CommentedConfig values) {
        ConfigGroup group = new ConfigGroup(id, groupName);
        for (UnmodifiableConfig.Entry entry : defaults.entrySet()) {
            Object entryValue = entry.getValue(), value = values.get(entry.getKey());
            if (entryValue instanceof /*? >=1.20.2 {*/ ModConfigSpec/*?} else {*/ /*ForgeConfigSpec*//*?}*/.ValueSpec spec) {
                Object defaultValue = spec.getDefault();
                try {
                    String translateKey = Objects.requireNonNullElseGet(spec.getTranslationKey(), entry::getKey);
                    ConfigBuilder<?, ?, ?> builder = this.process(values, TextUtil.translatableWithFallback(translateKey, TextFormatter.formatToTitleCase(translateKey)), entry, defaultValue, value, spec::test);
                    if (builder == null)
                        Jupiter.LOGGER.warn("Cannot find suitable entry for key={}, type={} in config={}:{}", entry.getKey(), defaultValue.getClass().getName(), this.modId, this.side);
                    else {
                        if (builder instanceof BaseEntry.Builder<?, ?, ?> baseBuilder && spec.getComment() != null)
                            baseBuilder.tooltip(spec.getComment());
                        group.add(builder.build());
                    }
                } catch (Exception e) {
                    Jupiter.LOGGER.error("Cannot load key={}, type={} in config={}:{}", entry.getKey(), defaultValue.getClass().getName(), this.modId, this.side, e);
                }
            } else if (entryValue instanceof UnmodifiableConfig spec && value instanceof CommentedConfig config) {
                Component name = TextUtil.translatableWithFallback(entry.getKey(), TextFormatter.formatToTitleCase(entry.getKey()));
                group.add(ConfigGroupEntry.builder(name, this.buildGroup(entry.getKey(), name, spec, config)).build());
            }
        }
        return group;
    }

    @SuppressWarnings({"rawtypes", "unchecked"})
    private ConfigBuilder<?, ?, ?> process(CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, Predicate<Object> validator) {
        AtomicReference<ConfigBuilder<?, ?, ?>> holder = new AtomicReference<>(null);
        //Simple
        this.processEntry(holder, values, name, entry, defaultValue, value, Boolean.class, BooleanEntry::builder);
        this.processEntry(holder, values, name, entry, defaultValue, value, Integer.class, IntegerEntry::builder);
        this.processEntry(holder, values, name, entry, defaultValue, value, Long.class, LongEntry::builder);
        this.processEntry(holder, values, name, entry, defaultValue, value, Double.class, DoubleEntry::builder);
        this.processEntry(holder, values, name, entry, defaultValue, value, String.class, StringEntry::builder);
        //Enum
        this.processEnum(holder, values, name, entry, defaultValue, value, defaultValue.getClass());
        //List
        if (Collection.class.isAssignableFrom(defaultValue.getClass()))
            //Some magic hack
            if (validator.test(List.of(false)))
                this.<Boolean, ListBooleanEntry, ListBooleanEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListBooleanEntry::builder);
            else if (validator.test(List.of(0)))
                this.<Integer, ListIntegerEntry, ListIntegerEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListIntegerEntry::builder);
            else if (validator.test(List.of(0L)))
                this.<Long, ListLongEntry, ListLongEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListLongEntry::builder);
            else if (validator.test(List.of(0D)))
                this.<Double, ListDoubleEntry, ListDoubleEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListDoubleEntry::builder);
            else if (validator.test(List.of("")))
                this.<String, ListStringEntry, ListStringEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListStringEntry::builder);
            else {//This method is unstable and usually failed to get
                Optional<?> any = ((List<?>) defaultValue).stream().findAny();
                if (any.isPresent() && any.get().getClass().isEnum())
                    this.processEnumCollection(holder, values, name, entry, defaultValue, value, (Enum) any.get());
                else {
                    Jupiter.LOGGER.warn("Notice: Jupiter cannot resolve empty List<Enum> since technical issue in Java, it is recommended to add a value in default value list.");
                    holder.set(SeparatorEntry.builder().text("jupiter.screen.cannot_process_list_enum").tooltip(name));
                }
            }
        return holder.get();
    }

    @SuppressWarnings("unchecked")
    private <T extends Enum<T>> void processEnum(AtomicReference<ConfigBuilder<?, ?, ?>> reference, CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, Class<?> clazz) {
        if (clazz.isEnum() && clazz.isAssignableFrom(defaultValue.getClass()) && value instanceof String valueStr) {
            Class<T> testClazz = (Class<T>) clazz;
            EnumEntry.Builder<T> builder = EnumEntry.builder(name, (T) defaultValue);
            builder.callback((v, r, d) -> values.set(entry.getKey(), v.name())).value(Enum.valueOf(testClazz, valueStr));
            reference.set(builder);
        }
    }

    @SuppressWarnings("unchecked")
    private <T, E extends IConfigEntry<T>, B extends ConfigBuilder<T, E, B>> void processEntry(AtomicReference<ConfigBuilder<?, ?, ?>> reference, CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, Class<T> clazz, BiFunction<Component, T, B> entryProvider) {
        if (clazz.isAssignableFrom(defaultValue.getClass()) && clazz.isAssignableFrom(value.getClass())) {
            B builder = entryProvider.apply(name, (T) defaultValue);
            builder.callback((v, r, d) -> values.set(entry.getKey(), v)).value((T) value);
            reference.set(builder);
        }
    }

    @SuppressWarnings("unchecked")
    private <T extends Enum<T>> void processEnumCollection(AtomicReference<ConfigBuilder<?, ?, ?>> reference, CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, T any) {
        Class<T> clazz = any.getDeclaringClass();
        ListEnumEntry.Builder<T> builder = ListEnumEntry.builder(name, ((List<T>) defaultValue), any);
        builder.callback((v, r, d) -> values.set(entry.getKey(), v.stream().map(Enum::name).toList())).value(new LinkedList<>(((List<String>) value).stream().map(x -> Enum.valueOf(clazz, x)).toList()));
        reference.set(builder);
    }

    @SuppressWarnings("unchecked")
    private <T, E extends IConfigEntry<List<T>>, B extends ConfigBuilder<List<T>, E, B>> void processCollectionEntry(AtomicReference<ConfigBuilder<?, ?, ?>> reference, CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, BiFunction<Component, List<T>, B> entryProvider) {
        B builder = entryProvider.apply(name, (List<T>) defaultValue);
        builder.callback((v, r, d) -> values.set(entry.getKey(), v)).value(new LinkedList<>((List<T>) value));
        reference.set(builder);
    }
}

==> jupiter-2.3.2-1.20.6-neoforge-sources.jar: com/iafenvoy/jupiter/compat/forgeconfigspec/NightConfigHolder.java <==
package com.iafenvoy.jupiter.compat.forgeconfigspec;

import com.electronwill.nightconfig.core.CommentedConfig;
import com.electronwill.nightconfig.core.UnmodifiableConfig;
import com.iafenvoy.jupiter.Jupiter;
import com.iafenvoy.jupiter.config.ConfigGroup;
import com.iafenvoy.jupiter.config.ConfigSide;
import com.iafenvoy.jupiter.config.entry.*;
import com.iafenvoy.jupiter.config.interfaces.ConfigBuilder;
import com.iafenvoy.jupiter.config.interfaces.ConfigEntry;
import com.iafenvoy.jupiter.util.TextFormatter;
import com.iafenvoy.jupiter.util.TextUtil;
import net.minecraft.network.chat.Component;
import net.minecraft.resources.ResourceLocation;
//? >= 1.20.2 {
import net.neoforged.neoforge.common.ModConfigSpec;
//?} else {
/*import net.minecraftforge.common.ForgeConfigSpec;
 *///?}

import java.util.*;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiFunction;
import java.util.function.Predicate;

//WARNING!!! DO NOT try to understand how these code work!!!
public final class NightConfigHolder {
    private final String modId;
    private final ConfigSide side;
    private final String fileName;
    private final UnmodifiableConfig defaults;
    private final CommentedConfig values;
    private final Runnable save;

    public NightConfigHolder(String modId, ConfigSide side, String fileName, UnmodifiableConfig defaults, CommentedConfig values, Runnable save) {
        this.modId = modId;
        this.side = side;
        this.fileName = fileName;
        this.defaults = defaults;
        this.values = values;
        this.save = save;
    }

    public ResourceLocation id() {
        return Jupiter.id(this.modId, this.side.name().toLowerCase(Locale.ROOT));
    }

    public Component title() {
        return TextUtil.literal(TextFormatter.formatToTitleCase(this.modId)).append(" ").append(TextUtil.translatable("jupiter.screen.%s_config".formatted(this.side.name().toLowerCase(Locale.ROOT))));
    }

    public ConfigSide getSide() {
        return this.side;
    }

    public String getFileName() {
        return this.fileName;
    }

    public void save() {
        this.save.run();
    }

    public List<ConfigGroup> toGroups() {
        return List.of(this.buildGroup(this.id().toString(), this.title(), this.defaults, this.values));
    }

    public ConfigGroup buildGroup(String id, Component groupName, UnmodifiableConfig defaults, CommentedConfig values) {
        ConfigGroup group = new ConfigGroup(id, groupName);
        for (UnmodifiableConfig.Entry entry : defaults.entrySet()) {
            Object entryValue = entry.getValue(), value = values.get(entry.getKey());
            if (entryValue instanceof /*? >=1.20.2 {*/ ModConfigSpec/*?} else {*/ /*ForgeConfigSpec*//*?}*/.ValueSpec spec) {
                Object defaultValue = spec.getDefault();
                try {
                    String translateKey = Objects.requireNonNullElseGet(spec.getTranslationKey(), entry::getKey);
                    ConfigBuilder<?, ?, ?> builder = this.process(values, TextUtil.translatableWithFallback(translateKey, TextFormatter.formatToTitleCase(translateKey)), entry, defaultValue, value, spec::test);
                    if (builder == null)
                        Jupiter.LOGGER.warn("Cannot find suitable entry for key={}, type={} in config={}:{}", entry.getKey(), defaultValue.getClass().getName(), this.modId, this.side);
                    else {
                        if (builder instanceof BaseEntry.Builder<?, ?, ?> baseBuilder && spec.getComment() != null)
                            baseBuilder.tooltip(spec.getComment());
                        group.addEntry(builder.build());
                    }
                } catch (Exception e) {
                    Jupiter.LOGGER.error("Cannot load key={}, type={} in config={}:{}", entry.getKey(), defaultValue.getClass().getName(), this.modId, this.side, e);
                }
            } else if (entryValue instanceof UnmodifiableConfig spec && value instanceof CommentedConfig config) {
                Component name = TextUtil.translatableWithFallback(entry.getKey(), TextFormatter.formatToTitleCase(entry.getKey()));
                group.addEntry(ConfigGroupEntry.builder(name, this.buildGroup(entry.getKey(), name, spec, config)).build());
            }
        }
        return group;
    }

    @SuppressWarnings({"rawtypes", "unchecked"})
    private ConfigBuilder<?, ?, ?> process(CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, Predicate<Object> validator) {
        AtomicReference<ConfigBuilder<?, ?, ?>> holder = new AtomicReference<>(null);
        //Simple
        this.processEntry(holder, values, name, entry, defaultValue, value, Boolean.class, BooleanEntry::builder);
        this.processEntry(holder, values, name, entry, defaultValue, value, Integer.class, IntegerEntry::builder);
        this.processEntry(holder, values, name, entry, defaultValue, value, Long.class, LongEntry::builder);
        this.processEntry(holder, values, name, entry, defaultValue, value, Double.class, DoubleEntry::builder);
        this.processEntry(holder, values, name, entry, defaultValue, value, String.class, StringEntry::builder);
        //Enum
        this.processEnum(holder, values, name, entry, defaultValue, value, defaultValue.getClass());
        //List
        if (Collection.class.isAssignableFrom(defaultValue.getClass()))
            //Some magic hack
            if (validator.test(List.of(false)))
                this.<Boolean, ListBooleanEntry, ListBooleanEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListBooleanEntry::builder);
            else if (validator.test(List.of(0)))
                this.<Integer, ListIntegerEntry, ListIntegerEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListIntegerEntry::builder);
            else if (validator.test(List.of(0L)))
                this.<Long, ListLongEntry, ListLongEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListLongEntry::builder);
            else if (validator.test(List.of(0D)))
                this.<Double, ListDoubleEntry, ListDoubleEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListDoubleEntry::builder);
            else if (validator.test(List.of("")))
                this.<String, ListStringEntry, ListStringEntry.Builder>processCollectionEntry(holder, values, name, entry, defaultValue, value, ListStringEntry::builder);
            else {//This method is unstable and usually failed to get
                Optional<?> any = ((List<?>) defaultValue).stream().findAny();
                if (any.isPresent() && any.get().getClass().isEnum())
                    this.processEnumCollection(holder, values, name, entry, defaultValue, value, (Enum) any.get());
                else {
                    Jupiter.LOGGER.warn("Notice: Jupiter cannot resolve empty List<Enum> since technical issue in Java, it is recommended to add a value in default value list.");
                    holder.set(SeparatorEntry.builder().text("jupiter.screen.cannot_process_list_enum").tooltip(name));
                }
            }
        return holder.get();
    }

    @SuppressWarnings("unchecked")
    private <T extends Enum<T>> void processEnum(AtomicReference<ConfigBuilder<?, ?, ?>> reference, CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, Class<?> clazz) {
        if (clazz.isEnum() && clazz.isAssignableFrom(defaultValue.getClass()) && value instanceof String valueStr) {
            Class<T> testClazz = (Class<T>) clazz;
            EnumEntry.Builder<T> builder = EnumEntry.builder(name, (T) defaultValue);
            builder.callback((v, r, d) -> values.set(entry.getKey(), v.name())).value(Enum.valueOf(testClazz, valueStr));
            reference.set(builder);
        }
    }

    @SuppressWarnings("unchecked")
    private <T, E extends ConfigEntry<T>, B extends ConfigBuilder<T, E, B>> void processEntry(AtomicReference<ConfigBuilder<?, ?, ?>> reference, CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, Class<T> clazz, BiFunction<Component, T, B> entryProvider) {
        if (clazz.isAssignableFrom(defaultValue.getClass()) && clazz.isAssignableFrom(value.getClass())) {
            B builder = entryProvider.apply(name, (T) defaultValue);
            builder.callback((v, r, d) -> values.set(entry.getKey(), v)).value((T) value);
            reference.set(builder);
        }
    }

    @SuppressWarnings("unchecked")
    private <T extends Enum<T>> void processEnumCollection(AtomicReference<ConfigBuilder<?, ?, ?>> reference, CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, T any) {
        Class<T> clazz = any.getDeclaringClass();
        ListEnumEntry.Builder<T> builder = ListEnumEntry.builder(name, ((List<T>) defaultValue), any);
        builder.callback((v, r, d) -> values.set(entry.getKey(), v.stream().map(Enum::name).toList())).value(new LinkedList<>(((List<String>) value).stream().map(x -> Enum.valueOf(clazz, x)).toList()));
        reference.set(builder);
    }

    @SuppressWarnings("unchecked")
    private <T, E extends ConfigEntry<List<T>>, B extends ConfigBuilder<List<T>, E, B>> void processCollectionEntry(AtomicReference<ConfigBuilder<?, ?, ?>> reference, CommentedConfig values, Component name, UnmodifiableConfig.Entry entry, Object defaultValue, Object value, BiFunction<Component, List<T>, B> entryProvider) {
        B builder = entryProvider.apply(name, (List<T>) defaultValue);
        builder.callback((v, r, d) -> values.set(entry.getKey(), v)).value(new LinkedList<>((List<T>) value));
        reference.set(builder);
    }
}
